use std::io::SeekFrom;
use nihav_core::codecs::*;
use nihav_core::demuxers::*;
use nihav_registry::detect;
use nihav_core::io::byteio::ByteReader;
use super::allreg::FullRegister;

pub struct RawStreamCtx<'a> {
    stream: NAStreamRef,
    pkt:    Box<dyn NAPacketiser + Send>,
    br:     &'a mut ByteReader<'a>,
    pts:    u64,
    seek:   SeekIndex,
}

impl<'a> RawStreamCtx<'a> {
    fn new(stream: NAStreamRef, packetiser: Box<dyn NAPacketiser + Send>, br: &'a mut ByteReader<'a>) -> Self {
        let mut seek = SeekIndex::new();
        seek.add_stream(0);
        Self { stream, pkt: packetiser, br, pts: 0, seek }
    }
    fn account_for_packet(&mut self, packet: &mut NAPacket) {
        let pos = self.br.tell() - (self.pkt.bytes_left() as u64);
        if packet.get_pts().is_none() && packet.get_duration().is_some() {
            packet.ts.pts = Some(self.pts);
        }
        if packet.is_keyframe() {
            let pts = packet.get_pts().unwrap_or(self.pts);
            let time = NATimeInfo::ts_to_time(pts, 1000, self.stream.tb_num, self.stream.tb_den);
            let in_range = if let Some(last) = self.seek.seek_info[0].entries.last() {
                    last.pts >= pts
                } else {
                    false
                };
            if !in_range {
                self.seek.add_entry(0, SeekEntry { time, pts, pos });
            }
        }
        self.pts += packet.get_duration().unwrap_or(0);
    }
    fn get_frame(&mut self) -> DemuxerResult<NAPacket> {
        let mut buf = [0; 1048576];
        loop {
            match self.pkt.get_packet(self.stream.clone()) {
                Ok(Some(mut packet)) => {
                    self.account_for_packet(&mut packet);
                    return Ok(packet);
                },
                Ok(None) => {},
                Err(DecoderError::ShortData) => {},
                _ => return Err(DemuxerError::InvalidData),
            };
            match self.br.read_buf_some(&mut buf) {
                Ok(size) => {
                    self.pkt.add_data(&buf[..size]);
                },
                Err(_) => {
                    match self.pkt.get_packet(self.stream.clone()) {
                        Ok(Some(mut packet)) => {
                            self.account_for_packet(&mut packet);
                            return Ok(packet);
                        },
                        Ok(None) | Err(DecoderError::ShortData) => return Err(DemuxerError::EOF),
                        _ => return Err(DemuxerError::InvalidData),
                    };
                },
            };
        }
    }
}

pub enum DemuxerObject<'a> {
    None,
    Normal(Demuxer<'a>),
    Raw(RawDemuxer<'a>, Vec<Option<Box<dyn NAPacketiser + Send>>>, bool),
    RawStream(RawStreamCtx<'a>),
}

impl<'a> DemuxerObject<'a> {
    pub fn create(br: &'a mut ByteReader<'a>, reg: &FullRegister, name: &str, is_raw: bool) -> DemuxerObject<'a> {
        if !is_raw {
            let res = detect::detect_format(name, br);
            let (dmx_name, _) = res.unwrap_or(("", detect::DetectionScore::No));
            if dmx_name != "" {
                println!("trying demuxer {} on {}", dmx_name, name);
                if let Some(dmx_fact) = reg.dmx_reg.find_demuxer(dmx_name) {
                    br.seek(SeekFrom::Start(0)).unwrap();
                    let dmx = create_demuxer(dmx_fact, br).unwrap();
                    return DemuxerObject::Normal(dmx);
                }
            }
            if dmx_name != "" {
                println!("trying raw demuxer {} on {}", dmx_name, name);
                if let Some(rdmx_fact) = reg.rdmx_reg.find_demuxer(dmx_name) {
                    br.seek(SeekFrom::Start(0)).unwrap();
                    let dmx = create_raw_demuxer(rdmx_fact, br).unwrap();
                    let mut pkts = Vec::new();
                    for stream in dmx.get_streams() {
                        if let Some(pcreate) = reg.pkt_reg.find_packetiser(stream.get_info().get_name()) {
                            let packetiser = (pcreate)();
                            pkts.push(Some(packetiser));
                        } else {
                            pkts.push(None);
                        }
                    }
                    return DemuxerObject::Raw(dmx, pkts, false);
                }
            }
            for rdmx in reg.rdmx_reg.iter() {
                if rdmx.check_format(br) {
                    println!("detected {} as {}", name, rdmx.get_name());
                    br.seek(SeekFrom::Start(0)).unwrap();
                    let dmx = create_raw_demuxer(*rdmx, br).unwrap();
                    let mut pkts = Vec::new();
                    for stream in dmx.get_streams() {
                        if let Some(pcreate) = reg.pkt_reg.find_packetiser(stream.get_info().get_name()) {
                            let packetiser = (pcreate)();
                            pkts.push(Some(packetiser));
                        } else {
                            pkts.push(None);
                        }
                    }
                    return DemuxerObject::Raw(dmx, pkts, false);
                }
            }
        }
        br.seek(SeekFrom::Start(0)).unwrap();
        let mut buf = vec![0; 1048576];
        let size = br.peek_buf(&mut buf).unwrap();
        let mut pname = "";

        for pinfo in reg.pkt_reg.iter() {
            let mut packetiser = (pinfo.get_packetiser)();
            packetiser.add_data(&buf[..size]);
            if packetiser.parse_stream(0).is_ok() {
                pname = pinfo.name;
                break;
            }
        }
        if pname != "" {
            println!("found raw stream of type {} for {}", pname, name);
            let pcreate = reg.pkt_reg.find_packetiser(pname).unwrap();
            let mut packetiser = (pcreate)();
            packetiser.add_data(&buf[..size]);
            let stream = packetiser.parse_stream(0).unwrap();
            packetiser.reset();
            DemuxerObject::RawStream(RawStreamCtx::new(stream, packetiser, br))
        } else {
            DemuxerObject::None
        }
    }
    pub fn is_none(&self) -> bool {
        match *self {
            DemuxerObject::None => true,
            _ => false,
        }
    }
    pub fn get_duration(&self) -> u64 {
        match *self {
            DemuxerObject::Normal(ref dmx) => dmx.get_duration(),
            DemuxerObject::Raw(ref dmx, _, _) => dmx.get_duration(),
            DemuxerObject::RawStream(ref ctx) => {
                NATimeInfo::ts_to_time(ctx.stream.duration, 1000, ctx.stream.tb_num, ctx.stream.tb_den)
            },
            _ => 0,
        }
    }
    pub fn get_num_streams(&self) -> usize {
        match *self {
            DemuxerObject::None => 0,
            DemuxerObject::Normal(ref dmx) => dmx.get_num_streams(),
            DemuxerObject::Raw(ref dmx, _, _) => dmx.get_num_streams(),
            DemuxerObject::RawStream(_) => 1,
        }
    }
    pub fn get_stream(&self, idx: usize) -> Option<NAStreamRef> {
        match *self {
            DemuxerObject::Normal(ref dmx) => dmx.get_stream(idx),
            DemuxerObject::Raw(ref dmx, _, _) => dmx.get_stream(idx),
            DemuxerObject::RawStream(ref ctx) if idx == 0 => Some(ctx.stream.clone()),
            _ => None,
        }
    }
    pub fn get_frame(&mut self) -> DemuxerResult<NAPacket> {
        match *self {
            DemuxerObject::Normal(ref mut dmx) => dmx.get_frame(),
            DemuxerObject::Raw(ref mut dmx, ref mut packetisers, ref mut eof) => {
                loop {
                    let mut has_some = false;
                    for (stream, p) in dmx.get_streams().zip(packetisers.iter_mut()) {
                        if let Some(ref mut pkts) = p {
                            match pkts.get_packet(stream.clone()) {
                                Ok(Some(pkt)) => return Ok(pkt),
                                Ok(None) | Err(DecoderError::ShortData) => {
                                    if *eof {
                                        *p = None;
                                    }
                                },
                                Err(err) => {
                                    println!("packetisation error {:?}", err);
                                    return Err(DemuxerError::InvalidData);
                                }
                            };
                            has_some |= p.is_some();
                        }
                    }
                    if !has_some {
                        return Err(DemuxerError::EOF);
                    }
                    if let Ok(data) = dmx.get_data() {
                        let id = data.get_stream().get_id();
                        for (i, stream) in dmx.get_streams().enumerate() {
                            if stream.get_id() == id {
                                if let Some(ref mut pkts) = packetisers[i] {
                                    pkts.add_data(&data.get_buffer());
                                }
                                break;
                            }
                        }
                    } else {
                        *eof = true;
                    }
                }
            },
            DemuxerObject::RawStream(ref mut ctx) => ctx.get_frame(),
            _ => unreachable!(),
        }
    }
    pub fn seek(&mut self, seek_time: NATimePoint) -> DemuxerResult<()> {
        match *self {
            DemuxerObject::Normal(ref mut dmx) => dmx.seek(seek_time),
            DemuxerObject::Raw(ref mut dmx, _, _) => dmx.seek(seek_time),
            DemuxerObject::RawStream(ref mut ctx) => {
                if seek_time == NATimePoint::None {
                    return Err(DemuxerError::SeekError);
                }
                if let Some(last) = ctx.seek.seek_info[0].entries.last() {
                    let in_index = match seek_time {
                            NATimePoint::None => unreachable!(),
                            NATimePoint::PTS(pts) => last.pts >= pts,
                            NATimePoint::Milliseconds(ms) => last.time >= ms,
                        };
                    if in_index {
                        if let Some(result) = ctx.seek.find_pos(seek_time) {
                            ctx.br.seek(SeekFrom::Start(result.pos))?;
                            ctx.pts = result.pts;
                            ctx.pkt.reset();
                            return Ok(());
                        }
                    }
                }
                if let Some(last) = ctx.seek.seek_info[0].entries.last() {
                    ctx.br.seek(SeekFrom::Start(last.pos))?;
                    ctx.pts = last.pts;
                    ctx.pkt.reset();
                }
                let mut key_pts = 0;
                while let Ok(pkt) = ctx.get_frame() {
                    if !pkt.ts.less_than(seek_time) && !pkt.ts.equal(seek_time) {
                        break;
                    }
                    if pkt.is_keyframe() {
                        key_pts = pkt.get_pts().unwrap_or(0);
                    }
                }
                let result = ctx.seek.find_pos(NATimePoint::PTS(key_pts)).unwrap();
                ctx.br.seek(SeekFrom::Start(result.pos))?;
                ctx.pts = result.pts;
                ctx.pkt.reset();
                Ok(())
            },
            _ => Err(DemuxerError::NotImplemented),
        }
    }
}

pub fn detect_tags(br: &mut ByteReader) -> (bool, u64, Option<u64>) {
    let mut is_raw = false;
    let mut start = 0;
    let mut end = None;

    // check for ID3v{2-4}
    let mut buf = [0; 5];
    br.peek_buf(&mut buf).unwrap();
    if &buf[0..3] == b"ID3" && buf[3] > 0 && buf[3] < 5 && buf[4] == 0 { //ID3 tag found, must be a raw stream
        br.read_skip(6).unwrap();
        let mut size = 0;
        for _ in 0..4 {
            let b = br.read_byte().unwrap();
            if (b & 0x80) != 0 {
                println!("Invalid ID3 size");
                break;
            }
            size = (size << 7) | u64::from(b);
        }
        start = size + 10;
        is_raw = true;
    }
    // check for ID3v1
    br.seek(SeekFrom::End(-128)).unwrap();
    let off = br.tell();
    br.peek_buf(&mut buf[..3]).unwrap();
    if &buf[0..3] == b"TAG" {
        end = Some(off);
    }
    // check for APETAG
    let mut buf = [0; 8];
    if let Some(off) = end {
        br.seek(SeekFrom::Start(off - 32)).unwrap();
    } else {
        br.seek(SeekFrom::End(-32)).unwrap();
    }
    let off = br.tell();
    br.read_buf(&mut buf).unwrap();
    if &buf == b"APETAGEX" {
        let ver     = br.read_u32le().unwrap();
        let size    = u64::from(br.read_u32le().unwrap());
        let _items  = br.read_u32le().unwrap();
        let flags   = br.read_u32le().unwrap();
        if ver == 1000 || (flags & 0x80000000) == 0 {
            end = Some(off - size + 32);
        } else {
            end = Some(off - size);
        }
    }

    (is_raw, start, end)
}
