use std::time::Duration;
use std::thread::JoinHandle;
use std::sync::{Arc, Mutex};
use std::sync::atomic::{AtomicBool, AtomicUsize, Ordering};
use std::sync::mpsc::{Receiver, SyncSender, TrySendError};
use std::str::FromStr;

use sdl2::AudioSubsystem;
use sdl2::audio::{AudioDevice, AudioCallback, AudioSpecDesired};

use nihav_core::formats::*;
use nihav_core::codecs::*;
use nihav_core::soundcvt::*;

use super::{DecoderStuff, PktSendEvent};

static SKIP_ADECODING: AtomicBool = AtomicBool::new(false);
static AUDIO_VOLUME: AtomicUsize = AtomicUsize::new(100);
static AUDIO_END: AtomicBool = AtomicBool::new(false);
static CUR_QUEUE_FILL: AtomicUsize = AtomicUsize::new(0);
static SAMPLE_RATE: AtomicUsize = AtomicUsize::new(0);
static CHANNELS: AtomicUsize = AtomicUsize::new(0);

static CURRENT_TIME: AtomicUsize = AtomicUsize::new(0);
static CURRENT_TIME_SET: AtomicBool = AtomicBool::new(false);

const QUEUE_INITIAL_SIZE: usize = 16384;
const QUEUE_REFILL_LIMIT: usize = 262144;

struct AudioQueue {
    queue:  Vec<i16>,
    start:  usize,
    end:    usize,
    srate:  usize,
    chans:  usize,
    spos:   usize,
}

impl AudioQueue {
    fn new(srate: usize, chans: usize) -> Self {
        SAMPLE_RATE.store(srate, Ordering::Relaxed);
        CHANNELS.store(chans, Ordering::Relaxed);
        CUR_QUEUE_FILL.store(0, Ordering::Relaxed);
        Self {
            queue:  Vec::with_capacity(QUEUE_INITIAL_SIZE),
            start:  0,
            end:    0,
            srate, chans,
            spos:   0,
        }
    }
    fn home(&mut self) {
        if self.start == 0 { return; }
        let fill = self.fill();
        if fill > 0 {
            if fill < self.start {
                let (dst, src) = self.queue.split_at_mut(self.start);
                dst[..fill].copy_from_slice(&src[..fill]);
            } else {
                for i in 0..fill {
                    self.queue[i] = self.queue[self.start + i];
                }
            }
        }
        self.start = 0;
        self.end = fill;
    }
    fn set_time(&mut self) {
        let fill = self.fill();
        let samp_time = self.spos.saturating_sub(fill / self.chans);
        CURRENT_TIME.store(samp_time * 1000 / self.srate, Ordering::Relaxed);
        CURRENT_TIME_SET.store(true, Ordering::Relaxed);
        CUR_QUEUE_FILL.store(self.fill(), Ordering::Relaxed);
    }
    fn add(&mut self, src: &[i16], samplepos: usize) {
        if self.end + src.len() > self.queue.len() {
            self.home();
        }
        if self.end + src.len() > self.queue.len() {
            self.queue.resize(self.end + src.len(), 0);
        }
        self.queue[self.end..][..src.len()].copy_from_slice(&src);
        self.end += src.len();
        self.spos = samplepos;
        self.set_time();
    }
    fn add_bytes(&mut self, src: &[u8], samplepos: usize) {
        let srclen = src.len() / 2;
        if self.end + srclen > self.queue.len() {
            self.home();
        }
        if self.end + srclen > self.queue.len() {
            self.queue.resize(self.end + srclen, 0);
        }
        for (dst, src) in self.queue[self.end..][..srclen].iter_mut().zip(src.chunks_exact(2)) {
            *dst = (u16::from(src[0]) + u16::from(src[1]) * 256) as i16;
        }
        self.end += srclen;
        self.spos = samplepos;
        self.set_time();
    }
    fn drain(&mut self, size: usize) {
        let fill = self.fill();
        if size >= fill {
            self.flush();
        } else {
            self.start += size;
        }
        self.set_time();
    }
    fn fill(&self) -> usize { self.end - self.start }
    fn flush(&mut self) {
        self.start  = 0;
        self.end    = 0;
    }
}

pub struct AudioOutput {
    queue:      Arc<Mutex<AudioQueue>>,
}

impl AudioCallback for AudioOutput {
    type Channel = i16;

    fn callback(&mut self, out: &mut [Self::Channel]) {
        let mut queue = self.queue.lock().unwrap();
        let dstlen = out.len();
        let copylen = queue.fill().min(dstlen);
        let volume = AUDIO_VOLUME.load(Ordering::Relaxed) as i32;
        if volume == 100 {
            out[..copylen].copy_from_slice(&queue.queue[queue.start..][..copylen]);
        } else {
            for (dst, &src) in out[..copylen].iter_mut().zip(queue.queue[queue.start..].iter()) {
                *dst = (i32::from(src) * volume / 100).max(-32768).min(32767) as i16;
            }
        }
        queue.drain(copylen);
        for el in out[copylen..].iter_mut() { *el = 0; }
    }
}

fn dummy_audio_thread(aprecv: Receiver<PktSendEvent>) -> JoinHandle<()> {
    std::thread::Builder::new().name("acontrol-dummy".to_string()).spawn(move ||{
                    loop {
                        match aprecv.recv() {
                            Ok(PktSendEvent::End) => break,
                            Ok(PktSendEvent::ImmediateEnd) => break,
                            Err(_) => {
                                break;
                            },
                            _ => {},
                        };
                    }
                }).unwrap()
}

type AudioPlaybackType = Option<AudioDevice<AudioOutput>>;

fn start_audio_decoding(asystem: &AudioSubsystem, ainfo: NAAudioInfo, mut audio_dec: DecoderStuff, aprecv: Receiver<PktSendEvent>) -> (AudioPlaybackType, JoinHandle<()>) {
    let ch = ainfo.channels.max(2);
    let desired_spec = AudioSpecDesired {
        freq:       Some(ainfo.sample_rate as i32),
        channels:   Some(ch),
        samples:    None
    };
    let dst_info = NAAudioInfo {
            sample_rate:    ainfo.sample_rate,
            channels:       ch,
            format:         SND_S16_FORMAT,
            block_len:      0,
        };
    let queue = Arc::new(Mutex::new(AudioQueue::new(ainfo.sample_rate as usize, ch as usize)));
    let qclone = queue.clone();
    let ret = asystem.open_playback(None, &desired_spec, |_spec| {
            AudioOutput {
                queue:      qclone,
            }
        });
    if ret.is_err() {
        return (None, dummy_audio_thread(aprecv))
    }
    let adevice = ret.unwrap();
    (Some(adevice), std::thread::Builder::new().name("acontrol".to_string()).spawn(move ||{
            let mut samplepos = 0usize;
            let dst_chmap = if dst_info.channels == 2 {
                    NAChannelMap::from_str("L,R").unwrap()
                } else {
                    NAChannelMap::from_str("C").unwrap()
                };
            SKIP_ADECODING.store(false, Ordering::Relaxed);
            loop {
                match aprecv.recv() {
                    Ok(PktSendEvent::Packet(pkt)) => {
                        loop {
                            if CUR_QUEUE_FILL.load(Ordering::Relaxed)
 < QUEUE_REFILL_LIMIT || SKIP_ADECODING.load(Ordering::Relaxed) {
                                break;
                            }
                            std::thread::sleep(Duration::from_millis(100));
                        }
                        if !SKIP_ADECODING.load(Ordering::Relaxed) {
                            if let Ok(frm) = audio_dec.dec.decode(&mut audio_dec.dsupp, &pkt) {
                                let buf = frm.get_buffer();
                                if let Some(pts) = frm.get_pts() {
                                    samplepos = NATimeInfo::ts_to_time(pts, u64::from(dst_info.sample_rate), frm.ts.tb_num, frm.ts.tb_den) as usize;
                                }
                                samplepos += buf.get_audio_length();
                                if let Ok(out_buf) = convert_audio_frame(&buf, &dst_info, &dst_chmap) {
                                    match out_buf {
                                        NABufferType::AudioI16(abuf) => {
                                            let mut qdata = queue.lock().unwrap();
                                            qdata.add(abuf.get_data(), samplepos);
                                            drop(qdata);
                                        },
                                        NABufferType::AudioPacked(abuf) => {
                                            let mut qdata = queue.lock().unwrap();
                                            qdata.add_bytes(abuf.get_data(), samplepos);
                                            drop(qdata);
                                        },
                                        _ => {},
                                    };
                                }
                            }
                        }
                    },
                    Ok(PktSendEvent::Flush) => {
                        audio_dec.dec.flush();
                        let mut qdata = queue.lock().unwrap();
                        qdata.flush();
                        SKIP_ADECODING.store(false, Ordering::Relaxed);
                    },
                    Ok(PktSendEvent::End) => break,
                    Ok(PktSendEvent::ImmediateEnd) => {
                        let mut qdata = queue.lock().unwrap();
                        qdata.flush();
                        break;
                    },
                    Ok(PktSendEvent::HurryUp) => {},
                    Err(_) => {
                        break;
                    },
                };
            }
            loop {
                let qdata = queue.lock().unwrap();
                if qdata.fill() == 0 || SKIP_ADECODING.load(Ordering::Relaxed) {
                    break;
                }
            }
            AUDIO_END.store(true, Ordering::Relaxed);
        }).unwrap())
}

pub struct AudioControl {
    aqueue:         Vec<PktSendEvent>,
    apsend:         SyncSender<PktSendEvent>,
    adevice:        AudioPlaybackType,
    athread:        JoinHandle<()>,
}

impl AudioControl {
    pub fn new(audio_dec: Option<DecoderStuff>, ainfo: Option<NAAudioInfo>, asystem: &AudioSubsystem) -> Self {
        let (apsend, aprecv) = std::sync::mpsc::sync_channel::<PktSendEvent>(20);
        let (adevice, athread) = if let Some(audio_dec) = audio_dec {
                start_audio_decoding(asystem, ainfo.unwrap(), audio_dec, aprecv)
            } else {
                (None, dummy_audio_thread(aprecv))
            };
        AUDIO_END.store(false, Ordering::Relaxed);

        Self {
            aqueue:     Vec::new(),
            apsend,
            adevice,
            athread,
        }
    }
    pub fn has_audio(&self) -> bool { self.adevice.is_some() }
    pub fn pause(&mut self) {
        if let Some(ref device) = self.adevice {
            device.pause();
        }
    }
    pub fn resume(&mut self) {
        if let Some(ref device) = self.adevice {
            device.resume();
        }
    }
    pub fn set_volume(&mut self, volume: usize) {
        AUDIO_VOLUME.store(volume, Ordering::Relaxed);
    }
    pub fn get_volume(&self) -> usize {
        AUDIO_VOLUME.load(Ordering::Relaxed)
    }
    pub fn is_audio_end(&self) -> bool {
        AUDIO_END.load(Ordering::Relaxed)
    }
    pub fn get_fill(&self) -> usize { CUR_QUEUE_FILL.load(Ordering::Relaxed) }
    pub fn get_time(&self) -> Option<u64> {
        if CURRENT_TIME_SET.load(Ordering::Relaxed) {
            Some(CURRENT_TIME.load(Ordering::Relaxed) as u64)
        } else {
            None
        }
    }
    pub fn get_time_left(&self) -> u64 {
        let srate = SAMPLE_RATE.load(Ordering::Relaxed);
        let chans = CHANNELS.load(Ordering::Relaxed);
        if srate != 0 && chans != 0{
            let fill = self.get_fill();
            (fill * 1000 / srate / chans) as u64
        } else {
            0
        }
    }

    pub fn get_queue_size(&self) -> usize { self.aqueue.len() }
    pub fn try_send_audio(&mut self, evt: PktSendEvent) -> bool {
        if self.aqueue.len() > 0 {
            self.aqueue.push(evt);
            false
        } else {
            self.try_send_event(evt)
        }
    }
    fn try_send_event(&mut self, evt: PktSendEvent) -> bool {
        if let Err(TrySendError::Full(evt)) = self.apsend.try_send(evt) {
            self.aqueue.insert(0, evt);
            false
        } else {
            true
        }
    }
    pub fn try_send_queued(&mut self) -> bool {
        while !self.aqueue.is_empty() {
            let pkt = self.aqueue.remove(0);
            if !self.try_send_event(pkt) {
                 return false;
            }
        }
        true
    }

    pub fn flush(&mut self) {
        self.pause();
        self.aqueue.clear();
        SKIP_ADECODING.store(true, Ordering::Release);
        CURRENT_TIME_SET.store(false, Ordering::Release);
        let _ = self.apsend.send(PktSendEvent::Flush);
    }
    pub fn finish(self) {
        SKIP_ADECODING.store(true, Ordering::Release);
        let _ = self.apsend.send(PktSendEvent::ImmediateEnd);
        self.athread.join().unwrap();
    }
}
