use std::thread::JoinHandle;
use std::sync::atomic::{AtomicBool, Ordering};
use std::sync::mpsc::{Receiver, SyncSender, TrySendError};
use std::thread;

use sdl2::render::Texture;

use nihav_core::frame::{NABufferType, NAVideoBuffer};
use nihav_core::formats::*;
use nihav_core::codecs::*;
use nihav_core::scale::*;

use super::{DecoderStuff, DispQueue, FrameRecord, PktSendEvent, FRAME_QUEUE_LEN};

static SKIP_VDECODING: AtomicBool = AtomicBool::new(false);
static VIDEO_END: AtomicBool = AtomicBool::new(false);

pub const FRAME_QUEUE_SIZE: usize = 25;

pub const SDL_RGB_FMT: NAPixelFormaton = NAPixelFormaton { model: ColorModel::RGB(RGBSubmodel::RGB), components: 3,
    comp_info: [
        Some(NAPixelChromaton { h_ss: 0, v_ss: 0, packed: true, depth: 8, shift: 0, comp_offs: 0, next_elem: 3 }),
        Some(NAPixelChromaton { h_ss: 0, v_ss: 0, packed: true, depth: 8, shift: 0, comp_offs: 1, next_elem: 3 }),
        Some(NAPixelChromaton { h_ss: 0, v_ss: 0, packed: true, depth: 8, shift: 0, comp_offs: 2, next_elem: 3 }),
        None, None
    ], elem_size: 3, be: false, alpha: false, palette: false };

pub struct VideoDecoder {
    yuv_pool:   NAVideoBufferPool<u8>,
    rgb_pool:   NAVideoBufferPool<u8>,
    tb_num:     u32,
    tb_den:     u32,
    dec:        DecoderStuff,
    ifmt:       NAVideoInfo,
    scaler:     NAScale,
    ofmt_rgb:   ScaleInfo,
    ofmt_yuv:   ScaleInfo,
    oinfo_yuv:  NAVideoInfo,
    oinfo_rgb:  NAVideoInfo,
}

impl VideoDecoder {
    pub fn new(width: usize, height: usize, tb_num: u32, tb_den: u32, dec: DecoderStuff) -> Self {
        let ofmt_rgb = ScaleInfo { width, height, fmt: SDL_RGB_FMT };
        let ofmt_yuv = ScaleInfo { width, height, fmt: YUV420_FORMAT };
        let oinfo_rgb = NAVideoInfo { width, height, flipped: false, format: SDL_RGB_FMT, bits: 24 };
        let oinfo_yuv = NAVideoInfo { width, height, flipped: false, format: YUV420_FORMAT, bits: 12 };
        Self {
            yuv_pool:   NAVideoBufferPool::new(FRAME_QUEUE_SIZE),
            rgb_pool:   NAVideoBufferPool::new(FRAME_QUEUE_SIZE),
            tb_num, tb_den,
            dec, ofmt_yuv, ofmt_rgb, oinfo_yuv, oinfo_rgb,
            scaler:     NAScale::new(ofmt_rgb, ofmt_rgb).unwrap(),
            ifmt:       NAVideoInfo { width: 0, height: 0, flipped: false, format: SDL_RGB_FMT, bits: 24 },
        }
    }
    fn convert_buf(&mut self, bt: NABufferType, ts: u64) -> Option<FrameRecord> {
        let vinfo = bt.get_video_info().unwrap();
        if self.ifmt.get_width() != vinfo.get_width() ||
            self.ifmt.get_height() != vinfo.get_height() ||
            self.ifmt.get_format() != vinfo.get_format() {
            self.ifmt = vinfo;
            let sc_ifmt = ScaleInfo { width: self.ifmt.get_width(), height: self.ifmt.get_height(), fmt: self.ifmt.get_format() };
            let do_yuv = if let ColorModel::YUV(_) = self.ifmt.get_format().get_model() { true } else { false };
            let ofmt = if do_yuv { self.ofmt_yuv } else { self.ofmt_rgb };
            self.scaler = NAScale::new(sc_ifmt, ofmt).unwrap();
        }
        let mut opic = if let ColorModel::YUV(_) = self.ifmt.get_format().get_model() {
                self.yuv_pool.prealloc_video(self.oinfo_yuv, 2).unwrap();
                while self.yuv_pool.get_free().is_none() {
                    if SKIP_VDECODING.load(Ordering::Relaxed) {
                        return None;
                    }
                    std::thread::yield_now();
                }
                NABufferType::Video(self.yuv_pool.get_free().unwrap())
            } else {
                self.rgb_pool.prealloc_video(self.oinfo_rgb, 0).unwrap();
                while self.rgb_pool.get_free().is_none() {
                    if SKIP_VDECODING.load(Ordering::Relaxed) {
                        return None;
                    }
                    std::thread::yield_now();
                }
                NABufferType::VideoPacked(self.rgb_pool.get_free().unwrap())
            };
        let ret = self.scaler.convert(&bt, &mut opic);
        if ret.is_err() { println!(" scaler error {:?}", ret.err()); return None; }
        ret.unwrap();
        let time = NATimeInfo::ts_to_time(ts, 1000, self.tb_num, self.tb_den);
        Some((opic, time))
    }
    pub fn next_frame(&mut self, pkt: &NAPacket) -> Option<FrameRecord> {
        if let Ok(frm) = self.dec.dec.decode(&mut self.dec.dsupp, pkt) {
            self.dec.reord.add_frame(frm);
            while let Some(frm) = self.dec.reord.get_frame() {
                let bt = frm.get_buffer();
                if let NABufferType::None = bt { continue; }
                let ts = frm.get_dts().unwrap_or_else(|| frm.get_pts().unwrap_or(0));
                return self.convert_buf(bt, ts);
            }
        }
        None
    }
    pub fn last_frame(&mut self) -> Option<FrameRecord> {
        while let Some(frm) = self.dec.reord.get_last_frames() {
            let bt = frm.get_buffer();
            if let NABufferType::None = bt { continue; }
            let ts = frm.get_dts().unwrap_or_else(|| frm.get_pts().unwrap_or(0));
            return self.convert_buf(bt, ts);
        }
        None
    }
    pub fn flush(&mut self) {
        self.dec.dec.flush();
        self.dec.reord.flush();
    }
}

fn start_video_decoding(width: usize, height: usize, tb_num: u32, tb_den: u32, video_dec: DecoderStuff, vprecv: Receiver<PktSendEvent>, vfsend: SyncSender<(NABufferType, u64)>) -> JoinHandle<()> {
    std::thread::Builder::new().name("vdecoder".to_string()).spawn(move ||{
            SKIP_VDECODING.store(false, Ordering::Relaxed);
            let mut vdec = VideoDecoder::new(width, height, tb_num, tb_den, video_dec);
            let mut skip_mode = FrameSkipMode::None;
            loop {
                match vprecv.recv() {
                    Ok(PktSendEvent::Packet(pkt)) => {
                        if !SKIP_VDECODING.load(Ordering::Relaxed) {
                            if let Some((buf, time)) = vdec.next_frame(&pkt) {
                                vfsend.send((buf, time)).unwrap();
                            }
                        }
                    },
                    Ok(PktSendEvent::Flush) => {
                        vdec.flush();
                        SKIP_VDECODING.store(false, Ordering::Relaxed);
                    },
                    Ok(PktSendEvent::End) => {
                        while vdec.yuv_pool.get_free().is_some() && vdec.rgb_pool.get_free().is_some() {
                            let ret = vdec.last_frame();
                            if ret.is_none() {
                                break;
                            }
                            vfsend.send(ret.unwrap()).unwrap();
                        }
                        VIDEO_END.store(true, Ordering::Relaxed);
                        break;
                    },
                    Ok(PktSendEvent::ImmediateEnd) => {
                        VIDEO_END.store(true, Ordering::Relaxed);
                        break;
                    },
                    Ok(PktSendEvent::HurryUp) => {
                        skip_mode = skip_mode.advance();
                        vdec.dec.dec.set_options(&[NAOption{
                                name: FRAME_SKIP_OPTION,
                                value: NAValue::String(skip_mode.to_string()),
                            }]);
                    },
                    Err(_) => {
                        break;
                    },
                };
            }
        }).unwrap()
}

trait Advance {
    fn advance(&self) -> Self;
}

impl Advance for FrameSkipMode {
    fn advance(&self) -> Self {
        match *self {
            FrameSkipMode::None             => FrameSkipMode::KeyframesOnly,
            FrameSkipMode::KeyframesOnly    => FrameSkipMode::IntraOnly,
            FrameSkipMode::IntraOnly        => FrameSkipMode::None,
        }
    }
}

fn output_yuv(yuv_texture: &mut Texture, buf: &NAVideoBuffer<u8>, width: usize, height: usize) {
    let src = buf.get_data();
    let ysstride = buf.get_stride(0);
    let ysrc = &src[buf.get_offset(0)..];
    let usstride = buf.get_stride(2);
    let usrc = &src[buf.get_offset(2)..];
    let vsstride = buf.get_stride(1);
    let vsrc = &src[buf.get_offset(1)..];
    yuv_texture.with_lock(None, |buffer: &mut [u8], pitch: usize| {
            let csize = pitch.min(width);
            for (dline, sline) in buffer.chunks_exact_mut(pitch).take(height).zip(ysrc.chunks_exact(ysstride)) {
                dline[..csize].copy_from_slice(&sline[..csize]);
            }
            let coff = pitch * height;
            let csize = (pitch / 2).min(width / 2);
            for (dline, sline) in buffer[coff..].chunks_exact_mut(pitch / 2).take(height/2).zip(vsrc.chunks(vsstride)) {
                dline[..csize].copy_from_slice(&sline[..csize]);
            }
            let coff = pitch * height + (pitch / 2) * (height / 2);
            for (dline, sline) in buffer[coff..].chunks_exact_mut(pitch / 2).take(height/2).zip(usrc.chunks(usstride)) {
                dline[..csize].copy_from_slice(&sline[..csize]);
            }
        }).unwrap();
}


pub struct VideoControl {
    vqueue:         Vec<PktSendEvent>,
    vpsend:         SyncSender<PktSendEvent>,
    vfrecv:         Receiver<FrameRecord>,
    do_yuv:         bool,
    vthread:        JoinHandle<()>,
}

impl VideoControl {
    pub fn new(video_dec: Option<DecoderStuff>, width: usize, height: usize, tb_num: u32, tb_den: u32) -> Self {
        let (vpsend, vprecv) = std::sync::mpsc::sync_channel::<PktSendEvent>(0);
        let (vfsend, vfrecv) = std::sync::mpsc::sync_channel::<FrameRecord>(FRAME_QUEUE_SIZE - 1);

        VIDEO_END.store(false, Ordering::Relaxed);

        let vthread = if let Some(video_dec) = video_dec {
                start_video_decoding(width, height, tb_num, tb_den, video_dec, vprecv, vfsend)
            } else {
                thread::Builder::new().name("vdecoder-dummy".to_string()).spawn(move ||{
                        loop {
                            match vprecv.recv() {
                                Ok(PktSendEvent::End) => break,
                                Ok(PktSendEvent::ImmediateEnd) => break,
                                Err(_) => {
                                    break;
                                },
                                _ => {},
                            };
                        }
                    }).unwrap()
            };


        Self {
            vqueue:         Vec::with_capacity(FRAME_QUEUE_LEN),
            vpsend, vfrecv,
            do_yuv:         false,
            vthread,
        }
    }
    pub fn flush(&mut self) {
        self.vqueue.clear();
        SKIP_VDECODING.store(true, Ordering::Release);
        for _ in 0..8 {
            let _ = self.vfrecv.try_recv();
        }
        let _ = self.vpsend.send(PktSendEvent::Flush);
        while self.vfrecv.try_recv().is_ok() { }
    }
    pub fn get_queue_size(&self) -> usize { self.vqueue.len() }
    pub fn is_filled(&self, size: usize) -> bool {
        self.vqueue.len() >= size
    }
    pub fn try_send_video(&mut self, evt: PktSendEvent) -> bool {
        if self.vqueue.len() > 0 {
            self.vqueue.push(evt);
            false
        } else {
            self.try_send_event(evt)
        }
    }
    fn try_send_event(&mut self, evt: PktSendEvent) -> bool {
        if let Err(TrySendError::Full(evt)) = self.vpsend.try_send(evt) {
            self.vqueue.insert(0, evt);
            false
        } else {
            true
        }
    }
    pub fn try_send_queued(&mut self) -> bool {
        while !self.vqueue.is_empty() {
            let pkt = self.vqueue.remove(0);
            if !self.try_send_event(pkt) {
                 return false;
            }
        }
        true
    }
    pub fn is_video_end(&self) -> bool {
        VIDEO_END.load(Ordering::Relaxed)
    }

    pub fn is_yuv(&self) -> bool { self.do_yuv }

    pub fn fill(&mut self, disp_queue: &mut DispQueue) {
        while !disp_queue.is_full() {
            let is_empty = disp_queue.is_empty();
            if let Ok((pic, time)) = self.vfrecv.try_recv() {
                let buf = pic.get_vbuf().unwrap();
                self.do_yuv = buf.get_info().get_format().get_model().is_yuv();
                let idx = disp_queue.end;
                disp_queue.move_end();
                let frm = &mut disp_queue.pool[idx];
                if !self.do_yuv {
                    let sstride = buf.get_stride(0);
                    let src = buf.get_data();
                    frm.rgb_tex.with_lock(None, |buffer: &mut [u8], pitch: usize| {
                            let csize = sstride.min(pitch);
                            for (dst, src) in buffer.chunks_mut(pitch).zip(src.chunks(sstride)) {
                                (&mut dst[..csize]).copy_from_slice(&src[..csize]);
                            }
                            true
                        }).unwrap();
                } else {
                    output_yuv(&mut frm.yuv_tex, &buf, disp_queue.width, disp_queue.height);
                }
                frm.valid = true;
                frm.is_yuv = self.do_yuv;
                frm.ts = time;
                if is_empty {
                    disp_queue.first_ts = time;
                }
                disp_queue.last_ts = time;
            } else {
                break;
            }
        }
    }

    pub fn finish(self) {
        SKIP_VDECODING.store(true, Ordering::Release);
        for _ in 0..8 {
            let _ = self.vfrecv.try_recv();
        }
        let _ = self.vpsend.send(PktSendEvent::ImmediateEnd);
        self.vthread.join().unwrap();
    }
}
